// This file is licensed under the Elastic License 2.0. Copyright 2021 StarRocks Limited.

package com.starrocks.sql.optimizer.statistics;

import com.google.common.base.Preconditions;

import java.util.Objects;

import static com.google.common.base.Preconditions.checkArgument;
import static java.lang.Double.NaN;
import static java.lang.Double.isFinite;
import static java.lang.Double.isInfinite;
import static java.lang.Double.isNaN;
import static java.lang.Math.max;
import static java.lang.Math.min;
import static java.util.Objects.requireNonNull;

// Calculate the cross range and ratio between column statistics
public class StatisticRangeValues {
    private final double low;
    private final double high;
    private final double distinctValues;

    public StatisticRangeValues(double low, double high, double distinctValues) {
        Preconditions.checkArgument(
                low <= high || (isNaN(low) && isNaN(high)),
                "low value must be less than or equal to high value or both values have to be NaN, got %s and %s respectively",
                low,
                high);
        this.low = low;
        this.high = high;

        checkArgument(distinctValues >= 0 || isNaN(distinctValues),
                "Distinct values count should be non-negative, got: %s", distinctValues);
        this.distinctValues = distinctValues;
    }

    public static StatisticRangeValues empty() {
        return new StatisticRangeValues(NaN, NaN, 0);
    }

    public boolean isEmpty() {
        return isNaN(low) && isNaN(high);
    }

    public boolean isBothInfinite() {
        return isInfinite(low) && isInfinite(high);
    }

    public static StatisticRangeValues from(ColumnStatistic column) {
        return new StatisticRangeValues(column.getMinValue(), column.getMaxValue(), column.getDistinctValuesCount());
    }

    public double getLow() {
        return low;
    }

    public double getHigh() {
        return high;
    }

    public double getDistinctValues() {
        return distinctValues;
    }

    public double length() {
        return this.high - this.low;
    }

    // Calculate the proportion of coverage between column statistic range
    public double overlapPercentWith(StatisticRangeValues other) {
        requireNonNull(other, "other is null");

        if (this.isEmpty() || other.isEmpty()) {
            return 0.0;
        }
        // If the low and high values is infinite, it represents either string type or unknown of column statistics.
        if (this.equals(other) && !isBothInfinite()) {
            return 1.0;
        }

        double lengthOfIntersect = min(this.high, other.high) - max(this.low, other.low);
        // lengthOfIntersect of char/varchar is infinite
        if (isInfinite(lengthOfIntersect)) {
            if (isFinite(this.distinctValues) && isFinite(other.distinctValues)) {
                return min(other.distinctValues / this.distinctValues, 1);
            }
            return StatisticsEstimateCoefficient.OVERLAP_INFINITE_RANGE_FILTER_COEFFICIENT;
        }
        if (lengthOfIntersect == 0) {
            // distinctValues equals 1 means the column statistics is unknown,
            // requires special treatment
            if (this.distinctValues == 1 && length() > 1) {
                return 0.5;
            }
            return 1 / max(this.distinctValues, 1);
        }
        if (lengthOfIntersect < 0) {
            return 0;
        }

        double length = length();
        if (isInfinite(length)) {
            return StatisticsEstimateCoefficient.OVERLAP_INFINITE_RANGE_FILTER_COEFFICIENT;
        }

        if (lengthOfIntersect > 0) {
            return lengthOfIntersect / length;
        }

        return NaN;
    }

    public StatisticRangeValues intersect(StatisticRangeValues other) {
        double newLow = max(low, other.low);
        double newHigh = min(high, other.high);
        if (newLow <= newHigh) {
            return new StatisticRangeValues(newLow, newHigh, min(this.distinctValues, other.distinctValues));
        }
        return empty();
    }

    public StatisticRangeValues union(StatisticRangeValues other) {
        double overlapPercentThis = this.overlapPercentWith(other);
        double overlapPercentOther = other.overlapPercentWith(this);
        double overlapNDVThis = overlapPercentThis * distinctValues;
        double overlapNDVOther = overlapPercentOther * other.distinctValues;
        double maxOverlapNDV = Math.max(overlapNDVThis, overlapNDVOther);
        double newNDV = maxOverlapNDV + ((1 - overlapPercentThis) * distinctValues) +
                ((1 - overlapPercentOther) * other.distinctValues);
        return new StatisticRangeValues(Math.min(low, other.low), Math.max(high, other.high), newNDV);
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) {
            return true;
        }
        if (o == null || getClass() != o.getClass()) {
            return false;
        }

        StatisticRangeValues that = (StatisticRangeValues) o;

        if (Double.compare(that.low, low) != 0) {
            return false;
        }
        if (Double.compare(that.high, high) != 0) {
            return false;
        }
        return Double.compare(that.distinctValues, distinctValues) == 0;
    }

    @Override
    public int hashCode() {
        return Objects.hash(low, high, distinctValues);
    }
}
